/*
 * id3 include stuff
 */

/* ====================================================================
 * Copyright (c) 2003 Brian Aker.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * 4. The name "libID3" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    brian@tangent.org.
 *
 * 5. Products derived from this software may not be called "libID3".
 *
 * 6. For any changes to the license please contact brian@tangent.org.
 *    He is quite willing to be flexible for you.
 *
 * THIS SOFTWARE IS PROVIDED BY BRIAN AKER ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL BRIAN AKER OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 */

//#include <math.h>
//#include <stdio.h>
//#include <stdlib.h>
//#include <string.h>
//#include <strings.h>
//#include <sys/mman.h>
//#include <sys/types.h>
//#include <sys/stat.h>
//#include <assert.h>
//#include <unistd.h>
//#include <errno.h>
//#include <fcntl.h>
//#include <ctype.h>

#define VERSION_SIZE 8

typedef long id3flags;
typedef struct ID3 ID3;

struct ID3 {
	char id3v1_version[VERSION_SIZE]; //It will never be this large!
	char id3v2_version[VERSION_SIZE]; //It will never be this large!
	char *filename; // Filename is only added when available, will be null terminated
	char *title;
	size_t title_size;
	char *artist;
	size_t artist_size;
	char *album;
	size_t album_size;
	char *year;
	size_t year_size;
	char *comment;
	size_t comment_size;
	char *track;
	byte track_number;
	size_t track_size;
	char *length_millis;
	size_t length_millis_size;
	byte *picture;
	size_t picture_size;
	char *genre_code;
	char genre[25];
	size_t genre_size;
	char *encoder;
	size_t encoder_size;
	char *blob;
	char signature[33];
	void (*processor)(ID3 *info, const char *name, const char *ptr, size_t length);
    char *data;

	// mp3 header info
	word mp3_version, mp3_layer, channels, bitrate, samplerate, duration;

	size_t data_size;      // Length of the actual tag
	void *passable;        // Void pointer you can use in your processor() callback function
	char *buffer;          // Block of memory used to hold tags
	char *ptr;             // Pointer in block of memory used to hold tags
	size_t length;         // Size of the data in ptr
	size_t size;           // Size of the ptr
	size_t tag_length;     // Length of the actual tag
	char user_memory;      // If the user supplied memory to use for ptr
	char allocated;        // If the ID3 object was allocated
	id3flags mask;
	id3flags mask_found;
};

static unsigned int bitrates[2][3][15] =
{
  /* MPEG-1 */
  {
    {0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448},
    {0, 32, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384},
    {0, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320}
  },
  /* MPEG-2 LSF, MPEG-2.5 */
  {
    {0, 32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 256},
    {0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160},
    {0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160}
  },
};

static unsigned int s_freq[3][4] =
{
  {44100, 48000, 32000, 0},
  {22050, 24000, 16000, 0},
  {11025, 8000, 8000, 0}
};

static unsigned int samples_per_frame[3][4] =
{
  {384, 384, 384, 0},
  {1152, 1152, 1152, 0},
  {1152, 576, 576, 0}
};

typedef struct {
	word version;
	word layer;
	word error_protection;
	word bitrate;
	word samplerate;
	word samples_per_frame;
	word padding;
	word extension;
	word channels;
	word frame_length;
	word duration;
	dword number_of_frames;
} mp3_header;

#define GENRE_MAX 148
#define ID3_INIT_SIZE 8192

#define isframeid(a) (isupper(a) || isdigit(a))
#define HUGE_STRING_LEN 8192

ID3 * create_ID3(ID3 *);
int destroy_ID3(ID3 *blob);
bool parse_file_ID3(ID3 *info, unsigned char *filename, bool read_art);
bool parse_blob_ID3(ID3 *info, byte *id3v1_blob, byte *id3v2_blob, size_t id3v1_blob_length, size_t id3v2_blob_length, bool read_art);
bool parse_mp3_header(byte *buffer, mp3_header *header);
void parse_mp3_frames(ID3 *info, byte *buffer, long file_size, long tag_length);
static bool mp3_check_vbr(byte* buffer, mp3_header *header) ;
static dword mp3_frame_length(mp3_header* header) ;
void set_memory_ID3(ID3 *info, char *ptr, size_t size);
void set_flags_ID3(ID3 *info, id3flags mask);
const char * genre_string(int genre);
//int ID3_to_file(unsigned char *filename, unsigned char *newfile);


// Error Codes
#define ID3_ERR_EMPTY_FILE 1
#define ID3_ERR_NO_TAG 2
#define ID3_ERR_UNSUPPORTED_FORMAT 3

// ID3 versions
#define ID3_VERSION_1 1
#define ID3_VERSION_2 2

#define TITLE_TAG        1              /* Title of the track*/
#define ARTIST_TAG       2              /* Performing Artist */
#define ALBUM_TAG        4              /* Album */
#define YEAR_TAG         8              /* Year for the track */
#define COMMENT_TAG     16              /* Comment */
#define TRACK_TAG       32              /* Track on the Album */
#define GENRE_TAG       64              /* Music Genre */
#define ENCODER_TAG    128              /* Encoder for the track */
#define VERSION_TAG    256              /* Version of the Tag */
#define KEEP_TAG       512              /* Encoder for the track */
#define SIGNATURE_TAG 1024              /* Create MD5 signature of the file */
#define KEEP_BLOB     2048              /* Keep the blob that the file used */
#define LENGTH_TAG	  4096              /* Track length */
#define PICTURE_TAG	  8192              /* Embedded album art */
#define VERSION1_TAG        (TITLE_TAG|ARTIST_TAG|ALBUM_TAG|YEAR_TAG|COMMENT_TAG|TRACK_TAG|GENRE_TAG|LENGTH_TAG|PICTURE_TAG)        /* Encoder for the track */
#define ALL_TAG        (TITLE_TAG|ARTIST_TAG|ALBUM_TAG|YEAR_TAG|COMMENT_TAG|TRACK_TAG|GENRE_TAG|ENCODER_TAG|VERSION_TAG|SIGNATURE_TAG|LENGTH_TAG|PICTURE_TAG)        /* Encoder for the track */

#define DEBUG 1
#ifdef DEBUG
#define DEBUG_ENTER(A) const char *DEBUG_FUNCTION=A;TAP_Print("Entering %s\n", DEBUG_FUNCTION);
#define DEBUG_RETURN_MESSAGE(A,B) {TAP_Print("Leaving %s: ", DEBUG_FUNCTION);TAP_Print(A);TAP_Print("\n");return B;}
#define DEBUG_RETURN_VOID TAP_Print("Leaving %s\n", DEBUG_FUNCTION);
#define DEBUG_PRINT(A) TAP_Print("\t%s: ", DEBUG_FUNCTION); TAP_Print(A);TAP_Print("\n");
#define DEBUG_RETURN(A) {TAP_Print("Leaving %s\n", DEBUG_FUNCTION);return A;}
#define WATCHPOINT TAP_Print("WATCHPOINT %s %d\n", __FILE__, __LINE__);
#else
#define DEBUG_ENTER(a);
#define DEBUG_RETURN(a); 
#define DEBUG_RETURN_MESSAGE(A,B);
#define DEBUG_RETURN_VOID;
#define DEBUG_PRINT(A);
#define DEBUG_ASSERT(A);
#define WATCHPOINT;
#endif
